
/* -*- mode: c++; c-basic-offset: 3; -*- */
#include "sensemonitor2.hh"
#include "jsonStack.hh"
#include "DVecType.hh"
#include "resampler.hh"
#include "MultiPipe.hh"
#include "FilterDesign.hh"
#include "FIRdesign.hh"
#include "FIRdft.hh"
#include "fSeries/PSD.hh"
#include "WelchPSD.hh"
#include "MeanMedianPSD.hh"
#include "window_api.hh"
#include "pipe_tools.hh"
#include "frame_name.hh"
#include "html/align.hh"
#include "html/color.hh"
#include "html/document.hh"
#include "html/font.hh"
#include "html/hline.hh"
#include "html/image.hh"
#include "html/link.hh"
#include "html/size.hh"
#include "html/table.hh"
#include "html/text.hh"
#include "html/writer.hh"
#include <vector>
#include <utility>
#include <cmath>

using namespace std;
using containers::fSeries;
using containers::PSD;
using namespace html;

EXECDAT(sense_monitor2)

//======================================  sense_monitor2 constructor
sense_monitor2::sense_monitor2(int argc, const char* argv[])
  : DatEnv(argc, argv), mOsc(getDacc()), mSampleRate(8192), mStride(60.0),
   mHighPass(8.0), mHighPassDf(2.0), mHtmlUpdateInterval(300.0),
   mEnableServer(true)
{
   try {
      configure(argv[argc - 1]);
      display(cout);
   } catch (std::exception& e) {
      finish();
      return;
   }

   //----------------------------------- Test that >0 rangers defined
   if (mRangers.empty()) {
      cerr << "sense_monitor2: No ranger instances defined" << endl;
      finish();
      return;
   }

   //-----------------------------------  Set up trend and monitor channels
   if (mChannel[2] != ':') {
      cerr << "sense_monitor2: Channel name does not start with 2-digit ifo"
	   << endl;
      finish();
      return;
   }
   string ifo = mChannel.substr(0,2);
   if (monitorName.empty()) monitorName = string("SenseMonitor_") + ifo;
   if (mEnableServer) {
      mServer.setServerName(monitorName.c_str());
   }

   mTrend.setIFO(ifo);
   mTrend.setName(monitorName);
   mTrend.setType(Trend::kMinute);
   if (!mTrendPath.empty()) mTrend.setFile(mTrendPath);
   for (auto pR=mRangers.begin(); pR !=mRangers.end();  pR++) {
      pR->bookTrends(mTrend);
      if (mEnableServer) pR->monitorSetup(mServer);
   }

   //-----------------------------------  Set up data access
   getDacc().setStride(mStride);
   getDacc().addChannel(mChannel);
   getDacc().setIgnoreMissingChannel(true);
   setStrideAlignment(long(mStride), 0);

   //-----------------------------------  Cumulative log stream
   if (!mCumulativeLog.empty()) {
      mCumLogStream.open(mCumulativeLog.c_str(), ios::app);
      if (mCumLogStream.is_open()) {
	 display(mCumLogStream);
	 mCumLogStream << endl;
	 mCumLogStream << "Time (GPS)  ";
	 for (auto pR = mRangers.begin(); pR != mRangers.end(); pR++) {
	    mCumLogStream << "  " << setw(10) << pR->refToolName();
	 }
	 mCumLogStream << endl;
      } else {
	 cerr << "sense_monitor2: Unable to open cumulative log file: "
	      << mCumulativeLog << endl;
      }
   }

   //-----------------------------------  More initialization.
   mNStride = 0;
}

//======================================  sense_monitor2 destructor
sense_monitor2::~sense_monitor2(void) {
}

//======================================  Handle an attention signal
void
sense_monitor2::Attention(void) {
   if (mEnableServer) mServer.Attention();
}

//======================================  Clear previous results.
void
sense_monitor2::clear(void) {
   for (auto pR = mRangers.begin(); pR != mRangers.end(); pR++) pR->clear();
}

//======================================  Configuration file json keys.
const char* global_keys[] = {
   "monitor_name", "channel", "osc-file", "osc-break-cond", "osc-gate-cond",
   "trend-file", "status-file", "resample", "highpass", "highpass-df",
   "prep-filter", "stride", "stride", "H0_Mpc", "ranges", "enable-server",
   "cumulative-log", 0
};

const char* ranger_keys[] = {
   "name", "monitor-default-prefix", "psd-monitor","integrand-monitor",
   "range-monitor", "redshift-monitor", "trend-channel", "redshift-trend",
   "history-length", "red-shift", "f-low", "f-high", "psd-estimate",
   "window", "window-length", "window-parameter", "overlap",
   "histo-channel", "histo-bins", "histo-low", "histo-high", 0
};

//======================================  Parse the configuration file.
void
sense_monitor2::configure(const std::string& file) {
   jsonStack js(file);
   string err_keys = js.validate(global_keys);
   if (!err_keys.empty()) {
      cerr << "sense_monitor2: Invalid keywords: " << err_keys << endl;
   }

   //-----------------------------------  Get channel, monitor name.
   js.fetch_data("monitor_name", monitorName);
   js.fetch_data("channel", mChannel);
   if (mChannel.empty()) {
      throw runtime_error("sense_monitor2: No channel specified");
   }
   mStatusFile = "$DMTHTMLOUT/index.html";
   js.fetch_data("status-file", mStatusFile);

   //-----------------------------------  Configure osc definitions.
   js.fetch_data("osc-file", mOscPath);
   js.fetch_data("osc-break-cond", mOscBreakCond);
   js.fetch_data("osc-gate-cond", mOscGateCond);
   js.fetch_data("trend-file", mTrendPath);

   //-----------------------------------  Set signal conditioning
   js.fetch_data("resample", mSampleRate);
   js.fetch_data("highpass", mHighPass);
   js.fetch_data("highpass-df", mHighPassDf);
   string filterdesign;
   js.fetch_data("prep-filter", mFilterDesign);
   js.fetch_data("stride", mStride);
   js.fetch_data("enable-server", mEnableServer);
   js.fetch_data("H0_Mpc", ranger::H0);

   //-----------------------------------  Cumulative log file
   const char* htmlout = (getenv("DMTHTMLOUT"));
   if (htmlout) mCumulativeLog = string(htmlout) + "/";
   mCumulativeLog += "Cumulative.log";
   js.fetch_data("cumulative-log", mCumulativeLog);

   //-----------------------------------  Parse ranger instance defintitions
   if (js.push_member("ranges")) {
      jsonStack::autopop range_array_pop(js);

      //--------------------------------  Iterate over ranger structures
      while (js.iterate()) {
	 jsonStack::autopop range_pop(js);

	 //-----------------------------------  Check the range keys
	 string err_rkeys = js.validate(ranger_keys);
	 if (!err_rkeys.empty()) {
	    cerr << "sense_monitor2: Invalid keywords: " << err_rkeys << endl;
	 }

	 //-----------------------------------  Name
	 string name;
	 if (!js.fetch_data("name", name)) {
	    cerr << "sense_monitor2: no name for tool" << endl;
	    continue;
	 }
	 ranger Ri(name);

	 //----------------------------  define trend and monitor channels
	 string chan_name;
	 if (js.fetch_data("monitor-default-prefix", chan_name)) {
	    Ri.monitorDefaultPrefix(chan_name);
	 }
	 if (js.fetch_data("psd-monitor", chan_name)) {
	    Ri.setPsdMonitor(chan_name);
	 }
	 if (js.fetch_data("integrand-monitor", chan_name)) {
	    Ri.setIntegrandMonitor(chan_name);
	 }
	 if (js.fetch_data("range-monitor", chan_name)) {
	    Ri.setRangeMonitor(chan_name);
	 }
	 if (js.fetch_data("trend-channel", chan_name)) {
	    Ri.setRangeTrend(chan_name);
	 }

	 //----------------------------- Set the ranger history length.
	 double history_length = 86400;
	 js.fetch_data("history-length", history_length);
	 Ri.setHistoryLength(history_length);

	 //-----------------------------  define red-shift channels
	 bool redshift = false;
	 js.fetch_data("red-shift", redshift);
	 Ri.enableRedShift(redshift);
	 if (redshift) {
	    if (js.fetch_data("redshift-monitor", chan_name)) {
	       Ri.setRedShiftMonitor(chan_name);
	    }
	    if (js.fetch_data("redshift-trend", chan_name)) {
	       Ri.setRedShiftTrend(chan_name);
	    }
	 }

	 //-----------------------------  Range histogram parameters
	 if (js.fetch_data("histo-channel", chan_name)) {
	    Ri.setRangeHistoName(chan_name);
	 }
	 int nBins(100);
	 double hlow(100), high(150);
	 if (js.fetch_data("histo-bins", nBins) +
	     js.fetch_data("histo-low", hlow) +
	     js.fetch_data("histo-high", high))
	 {
	    Ri.setHistogramBins(nBins, hlow, high);
	 }

	 //-----------------------------  Set analysis frequency range
	 double fmin=10.0, fmax=1400;
	 js.fetch_data("f-low", fmin);
	 js.fetch_data("f-high", fmax);
	 Ri.setFmin(fmin);
	 Ri.setFmax(fmax);

	 //-----------------------------  Set up the psd estimator
	 string psd_type("welch"), window("tukey");
	 double window_length = 4.0, window_param = 0.75;
	 js.fetch_data("psd-estimate", psd_type);
	 js.fetch_data("window", window);
	 js.fetch_data("window-length", window_length);
	 js.fetch_data("window-parameter", window_param);
	 unique_ptr<psd_estimate> estimate;
	 if (psd_type == "welch") {
	    estimate.reset(new WelchPSD());
	 } else if (psd_type == "mean-median") {
	    estimate.reset(new MeanMedianPSD());
	 } else {
	    throw runtime_error("invalid psd-estimate type");
	 }
	 unique_ptr<window_api> win_instance;
	 try {
	    win_instance.reset(window_factory_1(window, 0, window_param));
	 } catch (runtime_error& e) {
	 }
	 if (!win_instance) try {
	    win_instance.reset(window_factory(window, 0));
	 } catch (runtime_error& e) {
	    cerr << "Error in range: " << name
		 << " configuration: undefined window type: " << window << endl;
	 }
	 estimate->set_stride(window_length);
	 estimate->set_window(*win_instance, true);
	 double overlap = 0;
	 if (js.fetch_data("overlap", overlap)) {
	    estimate->set_overlap(overlap);
	 }
	 if (psd_type == "mean-median") {
	    MeanMedianPSD* pMME = dynamic_cast<MeanMedianPSD*>(estimate.get());
	    pMME->set_averages(mStride/(window_length * (1-pMME->overlap())));
	 }
	 estimate->status(cout);
	 Ri.setPSDTool(*estimate);

	 //-----------------------------  Define GW template & sigma threshold
	 string gw_type = "bns-1.4x1.4";
	 js.fetch_data("gw-type", gw_type);
	 Ri.setGWType(gw_type);

	 double threshold = 8.0;
	 js.fetch_data("threshold-sigma", threshold);
	 Ri.setThreshold(threshold);

	 //-----------------------------  F-domain calibration filter
	 string fd_design;
	 js.fetch_data("fd-calibration", fd_design);

	 //-----------------------------  Add ranger to the list
	 mRangers.push_back(std::move(Ri));
      }
   }

   //-----------------------------------  Read osc files
   if (!mOscPath.empty()) mOsc.readConfig(mOscPath);

   //-----------------------------------  set up resample & conditioning filters
   mResampler.set(resampler(mSampleRate));
   set_fir_mode(mResampler.get(), FIRdft::fm_causal);

   if (mHighPass > 0) {
      FIRdft hp_filter(dFirW(0, mSampleRate, "Kaiser", "HighPass",
			     mHighPass, 0, 100.0, mHighPassDf));
      mConditioner.set(hp_filter);
   }
   if (!mFilterDesign.empty()) {
      FilterDesign fd(mFilterDesign.c_str(), mSampleRate);
      if (mConditioner.null()) {
	 mConditioner.set(fd.release());
      } else {
	 MultiPipe* mp = new MultiPipe;
	 mp->addPipe(mConditioner.release());
	 mp->addPipe(fd.release());
	 mConditioner.set(mp);
      }
   }
   set_fir_mode(mConditioner.get(), FIRdft::fm_causal);
}

//======================================  Display the configuration
void
sense_monitor2::display(ostream& out) const {
   out << "sense_monitor2 configuration" << endl;
   out << "  Channel name:         " << mChannel << endl;
   out << "  OSC file path:        " << mOscPath << endl;
   out << "  OSC break condition:  " << mOscBreakCond << endl;
   out << "  OSC gate condition:   " << mOscGateCond << endl;
   out << "  Trend file path:      " << mTrendPath << endl;
   out << "  Sample rate:          " << mSampleRate << endl;
   out << "  Analysis stride:      " << mStride << endl;
   out << "  Highpass frequency:   " << mHighPass << endl;
   out << "  HighPass transition:  " << mHighPassDf << endl;
   out << "  Conditioning filter:  " << mFilterDesign << endl;
   Interval delay(0.0);
   if (!mResampler.null())   delay  = mResampler->getTimeDelay();
   if (!mConditioner.null()) delay += mConditioner->getTimeDelay();
   out << "  Conditioning delay:   " << double(delay) << endl;
   out << "Ranger instances: " << endl;
   for (auto pR = mRangers.begin(); pR != mRangers.end(); pR++) {
      out << endl;
      pR->display(out);
   }
}

//======================================  Process a data stride.
void
sense_monitor2::ProcessData(void) {
   TSeries ts = *(getDacc().refData(mChannel));
   if (!mStartTime) {
      mStartTime = ts.getStartTime();
   }

   try {
      mResampler.dataCheck(ts);
      clear();
   } catch (exception& e) {
      reset();
      cerr << "sense_monitor2: Signal conditioning reset at gps: "
	   << ts.getStartTime() << endl;
   }

   TSeries inData = mConditioner(mResampler(ts));
   if (Debug() > 1) inData.dump_header(cout);

   //-----------------------------------  Extend monitor history to fill time
   Time     t0 = inData.getStartTime();
   Interval dt = inData.getInterval();
   for (auto pR = mRangers.begin(); pR != mRangers.end(); pR++) {
      pR->padHistory(t0, dt);
   }

   //-----------------------------------  calculate ranges if data are valid
   if (!mOsc.defined(mOscBreakCond) || mOsc.satisfied(mOscBreakCond)) {
      for (auto pR = mRangers.begin(); pR != mRangers.end(); pR++) {
	 pR->calculate(inData);
	 pR->fillTrends(mTrend);
      }
      mNStride++;
   }

   //-----------------------------------  Update trends
   Time tEnd = getDacc().getCurrentTime();
   mTrend.Update(tEnd);

   //-----------------------------------  Update Cumulative log file
   if (mCumLogStream.is_open()) {
      mCumLogStream << t0.getS() << "  " << fixed << setw(10)
		    << setprecision(2);
      for (auto pR = mRangers.begin(); pR != mRangers.end(); pR++) {
	 mCumLogStream << "  " << pR->range();
      }
      mCumLogStream << endl;
   }

   //-----------------------------------  Write status page.
   if (!mLastHtmlUpdate || mLastHtmlUpdate + mHtmlUpdateInterval <= tEnd) {
      put_html(tEnd);
      mLastHtmlUpdate = tEnd;
   }
}

//======================================  Print the html Status page
const char*
kHtmlTitle("sensemonitor2 (version 1.0): Multi-template GW Sensitivity Monitor");
const char* kHtmlLIGO("Laser Interferometer Gravitational-wave Observatory");
const char* const kHtmlLIGOLink("http://www.ligo.caltech.edu");

void
sense_monitor2::put_html(const Time& t) const {
   Time next = t + mHtmlUpdateInterval;
   const char* time_fmt = "%M %d, %Y at %H:%N:%S %Z";
   html::font fonttitle("Helvetica");
   html::document doc (kHtmlTitle);
   html::block titleblk ("center");
   titleblk.lineBreak();
   titleblk << (text(kHtmlTitle) << fonttitle << color(0, 0, 128) << html::size(+4));
   titleblk.lineBreak();
   titleblk << html::text ("&nbsp;");
   doc << titleblk
       << html::hline()
       << (html::text("This monitor calculates the approximate range at which ")
	   << "the interferometer can detect a specified gravitational "
	   << "wave source above a specified SNR threshold")
       << html::hline();
   //---------- Last update time.
   html::block firstupdateblk("center");
   firstupdateblk << text("This page was last updated: ") << text(t, time_fmt);
   firstupdateblk.lineBreak();
   firstupdateblk << text("Next scheduled update time: ")
		  << text(next, time_fmt);
   firstupdateblk.lineBreak();
   doc.addObject (firstupdateblk);
   doc.addObject (html::hline());     //---------- Run parameters
   html::block run_param_blk ("center");
   run_param_blk.lineBreak();
   run_param_blk << (text("Run Parameters") << fonttitle << html::size(+2));
   run_param_blk.lineBreak();     //---------- Parameters table.
   html::table params;     //----- WARNING:  Must make columns first, then rows.
   params.addColumn("");
   params.addColumn("");
   for (int j=0; j<11; ++j) {
      params.addRow();
   }

   params.refCell(0, 0) << "Channel: " << html::align("right");
   params.refCell(0, 1) << mChannel << html::align("left");
   params.refCell(1, 0) << "Stride: " << html::align("right");
   params.refCell(1, 1) << (text(mStride) << " sec") << html::align("left");
   params.refCell(2, 0) << "Strides Processed: " << html::align("right");
   params.refCell(2, 1) << text(mNStride+1) << html::align("left");
   params.refCell(3, 0) << "Highpass filter cutoff : " << html::align("right");
   params.refCell(3, 1) << (text(mHighPass) << " Hz") << html::align("left");
   params.refCell(4, 0) << "Monitor Start Time (GPS): " << html::align("right");
   params.refCell(4, 1) << text(Time(mStartTime), "%s") << html::align("left");
   params.refCell(5, 0) << "(UTC): " << html::align("right");
   params.refCell(5, 1) << text(Time(mStartTime), time_fmt) << html::align("left");
   run_param_blk.addObject(params);
   doc.addObject (run_param_blk);
   //---------- Current status section.
   html::block status_blk ("center");
   status_blk.lineBreak();
   status_blk << (text("Current Status") << fonttitle << html::size(+2));
   status_blk.lineBreak();

   html::table results;
   results.addColumn("GPS Time");
   results.addColumn("Number of strides processed");
   results.addColumn("Next Update Time (UTC)");
   results.addRow();
   results.refCell(0, 0) << text(t, "%s") << html::align("right");
   results.refCell(0, 1) << text(mNStride) << html::align("right");
   results.refCell(0, 2) << text(next) << html::align("right");
   results.setBorder(true);
   status_blk.addObject(results);

   //-----------------------------------  Range table
   status_blk.lineBreak();
   status_blk << (text("Current Range") << fonttitle << html::size(+2));
   status_blk.lineBreak();

   html::table ranges;
   ranges.addColumn("Tool name");
   ranges.addColumn("Source");
   ranges.addColumn("PSD tool");
   ranges.addColumn("Window");
   ranges.addColumn("Frequency band");
   ranges.addColumn("Range (1m)");
   ranges.addColumn("Range (5m)");
   ranges.addColumn("Range (15m)");
   for (auto pR=mRangers.begin(); pR != mRangers.end(); pR++) {
      ranges.addRow();
      size_t i = ranges.getNRow()-1;
      ranges.refCell(i, 0) << pR->refToolName() << html::align("left");
      ranges.refCell(i, 1) << pR->refGWType() << html::align("left");
      ranges.refCell(i, 2) << pR->psd_tool_type() << html::align("left");
      ranges.refCell(i, 3) << pR->psd_window_type() << html::align("left");
      ranges.refCell(i, 4) << text(pR->fmin()) << "-" << text(pR->fmax());
      ranges.refCell(i, 5) << text(pR->recent(mStride)) << html::align("right");
      ranges.refCell(i, 6) << text(pR->recent(5*mStride)) << html::align("right");
      ranges.refCell(i, 7) << text(pR->recent(15*mStride)) << html::align("right");
   }
   ranges.setBorder(true);
   status_blk.addObject(ranges);
   doc.addObject(status_blk);

   //---------- Insert error message link
   html::block errorinfo ("center");
   html::link errormessage ("Error Messages", "Error.log");
   errorinfo.addObject (errormessage);
   errorinfo.lineBreak();
   doc.addObject (errorinfo);

   //---------- Insert links to current (this run of monitor) and
   //           cumulative (all runs) log files.
   html::block output ("center");
   html::link outputlink1 ("Monitor Output: more details", "Range.log");
   output.addObject (outputlink1);
   output.lineBreak();

   html::link outputlink2("Cumulative Monitor Output from all runs",
			  "Cumulative.log");
   output.addObject (outputlink2);
   output.lineBreak();
   output.lineBreak();
   doc.addObject (output);

   //---------- Last update time.
   doc.addObject (html::hline());
   html::block updateblk ("center");
   updateblk << "This page was last updated: "
	     << TimeString (t, "%M %d, %Y at %H:%N:%S %Z");
   updateblk.lineBreak();

   //---------- Author info.
   updateblk << "This monitor was written by "
	     << html::link("John Zweizig", "mailto:john.zweizig@ligo.org")
	     << " based on SenseMonitor written by Patrick Sutton";
   doc.addObject (updateblk);
   doc.addObject (html::hline());

   //---------- LIGO logo.
   html::block logoblk ("div");
   logoblk.addAttr ("align", html::align ("right"));
   logoblk.addObject (html::link (kHtmlLIGO, kHtmlLIGOLink));
   html::image logo;
   logo.setSource ("ligo_logo.gif");
   logo.setWidth ("80");
   logoblk.addObject (logo);
   doc.addObject (logoblk);

   //---------- Write web page to file.
   string file = frame_name::resolve_env(mStatusFile);
   ofstream out (file.c_str());
   if (out) {
      html::writer htmlout (out);
      doc.write (htmlout);
   }
}

//=====================================  Reset monitor processing
void
sense_monitor2::reset(void) {
   mResampler.reset();
   mConditioner.reset();
   for (auto pR = mRangers.begin(); pR != mRangers.end(); pR++) {
      pR->reset();
   }
}
